#==============================================================================
#
# Copyright (c) 2013 Campanja AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# A copy of the license is included in the file LICENSE.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#==============================================================================

require 'erb'
require 'rubygems'
require 'json'
require 'thread'

module Fluent
  class RouterOutput < Fluent::Output

    Fluent::Plugin.register_output('router', self)
    MIN_REAPER_INTERVAL = 15 #sec
    ROUTER_STATE_FILE = "router_state.json"

    def initialize
      super
      @outputs = {}
      @last_used_time = {}
      @last_reaper_run = Time.now
      @semaphore = Mutex.new
    end

    def configure(conf)
      super
      log "Starting the out_router plugin"
      @inactivity_timeout = conf['inactivity_timeout'].to_i
      @output_config = conf.elements.select { |e| e.name == 'config' }.first
      mark_used @output_config
      filename = File.join(File.dirname(Fluent::DEFAULT_CONFIG_PATH),
                           ROUTER_STATE_FILE)
      @state_serializer = ConsistentArrayStorage.new(filename)
      output_keys = @state_serializer.load

      # Restarts all output plugins for each stored key after a crash
      # or a stop.
      # This way we ensure every needed buffered output plugin, like
      # the s3 plugin, will start after a crash and flush its buffer
      # files to s3.
      log "Reading file " + filename +
        " and found output keys: " + output_keys.to_json
      output_keys.each do |key|
        start_output(key)
      end
    end

    def emit(tag, es, chain)
      # This is the router logic. Splitting tag on character '.'
      # uses the value after '.' as the router key.
      key = tag.split(/\./)[1]
      out = @outputs[key]
      unless out
        out = start_output(key)
      end
      out.emit tag, es, chain
      now = Time.now
      @last_used_time[key] = now
      # Removed call to reaper because of two issues:
      # https://github.com/campanja/fluent-output-router/issues/3
      # https://github.com/campanja/fluent-output-router/issues/4
      #run_reaper(now)
    end

    private

    # start_output is synchroned with a mutex
    # to avoid any concurrency problems if the input plugin is multithreaded
    # for example: the scribe input plugin is multitreaded by default
    def start_output(key)
      @semaphore.synchronize do
        unless @outputs[key]
          log "Starting a new output of type " + @output_config['type'] +
            " for key #{key}"
          @outputs[key] = Fluent::Plugin.new_output @output_config['type']
          @last_used_time[key] = Time.now
          config = @output_config.clone
          template! config, :key => key
          out = @outputs[key]
          out.configure config
          out.start
          @state_serializer.store(@outputs.keys)
        end
        @outputs[key]
      end
    end

    def stop_output(key)
      log "Stopping output of type " + @output_config['type'] +
        " for key #{key} due to inactivity (no data comming in)"
      out = @outputs.delete(key)
      @last_used_time.delete(key)
      out.shutdown
      @state_serializer.store(@outputs.keys)
    end

    def mark_used(conf)
      conf.used = conf.keys
      conf.elements.each { |e| mark_used e }
    end

    def template!(conf, keys)
      conf.each do |k,v|
        case v
        when Hash
          conf[k] = template v, keys
        when String
          conf[k] = ERB.new(v).result(binding)
        else
          $log.error "Template error #{v.inspect}"
          $log.error_backtrace
        end
      end
    end

    def log(string)
      $log.info "out_router: " + string
    end

    # This is the reaper that is used to avoids resource leaks
    # It will stop any active output plugin that does not get any input
    # The code in run_reaper is synchroned with a mutex
    # to avoid any concurrency problems if the input plugin is multithreaded
    # for example: the scribe input plugin is multitreaded by default
    def run_reaper(now)
      if now - @last_reaper_run > MIN_REAPER_INTERVAL then
        @last_reaper_run = now
        @semaphore.synchronize do
          @last_used_time.each {|key, stored_time|
            if now - stored_time > @inactivity_timeout then
              stop_output(key)
            end }
        end
      end
    end
  end

    # ConsistentArrayStorage serializes/deserializes an array to disc.
    # It uses two files.
    # First it writes the serialized array into a tmp file and then renames
    # the tmp file to the correct filename.
    # The rename action is fast and Linux gives a guarantee that
    # the rename operation is atomic (on local discs). This removes the risk
    # of any kind of crash leaving the @store_file file in a bad state.
    class ConsistentArrayStorage
      def initialize(filename)
        @store_file = filename
        @temp_file = filename + ".tmp"
      end

      def store(array)
        file = File.open(@temp_file, 'w')
        file << array.to_json
        file.close
        File.rename(@temp_file, @store_file)
      end

      #Returns the array stored on disc.
      #Returns an empty array if the file does not exist
      def load
        if File.exists?(@store_file) then
          file = File.open(@store_file, 'r')
          json_string = file.read
          file.close
          JSON.parse(json_string)
        else
          []
        end
      end
    end

end
